;; Generated package description from simple-httpd.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "simple-httpd" "1.6.0.20260519.102820" "Pure Elisp HTTP server" '((emacs "27.1") (compat "31")) :commit "5a9c69ae93ed5e2cdca98f51b41f6a6782091d74" :authors '(("Christopher Wellons" . "wellons@nullprogram.com")) :maintainer '(("Philip Kaludercic" . "philipk@posteo.net") ("Daniel Mendler" . "mail@daniel-mendler.de")) :keywords '("network" "comm") :url "https://github.com/skeeto/emacs-http-server")
